// lsqr_solve_lapack.cc -*-c++-*-

#include<lsqr_solve_lapack.h>

#include <externals_config_0.h>

#include <boost/numeric/bindings/traits/type_traits.hpp>
#include <boost/numeric/bindings/traits/traits.hpp>
#include <boost/numeric/bindings/traits/detail/utils.hpp>

#include<cassert>
#include<iostream>
#include<cstdlib>

#include<cstddef> // for NULL

#define LAPACK_DGELS PR1_FORTRAN_NAME(dgels)
#define LAPACK_ZGELS PR1_FORTRAN_NAME(zgels)

extern "C" {
    void LAPACK_DGELS( char const * const trans,
		       int const * const m, int const * const n,
		       int const * const nrhs,
		       double * const a, int const * const lda,
		       double * const b, int const * const ldb,
		       double * const work, int const * const lwork,
		       int * info );
    void LAPACK_ZGELS( char const * const trans,
		       int const * const m, int const * const n,
		       int const * const nrhs,
		       dcomplex_t * const a, int const * const lda,
		       dcomplex_t * const b, int const * const ldb,
		       dcomplex_t * const work, int const * const lwork,
		       int * info );
}

namespace {
    namespace traits = boost::numeric::bindings::traits ;
}

namespace pu {

    inline int const
    gels_templ_diver_raw( bool const transp,
			  int const m, int const n, int const nrhs,
			  double * const A, int const LDA,
			  double * const b, int const LDB,
			  double * const work, int const lwork ){
	int info(0);
	char const transp_char( transp ? 'T' : 'N' );
	LAPACK_DGELS( &transp_char, &m, &n, &nrhs,
		      A, &LDA,
		      b, &LDB,
		      work, &lwork,
		      &info );
	return info;
    }

    inline int const
    gels_templ_diver_raw( bool const transp,
			  int const m, int const n, int const nrhs,
			  traits::complex_d * const A, int const LDA,
			  traits::complex_d * const b, int const LDB,
			  traits::complex_d * const work, int const lwork ){
	int info(0);
	char const transp_char( transp ? 'C' : 'N' );
	LAPACK_ZGELS( &transp_char, &m, &n, &nrhs,
		      traits::complex_ptr (A), &LDA,
		      traits::complex_ptr (b), &LDB,
		      traits::complex_ptr (work), &lwork,
		      &info );
	return info;
    }

    template<typename MATA, typename VTYPE, typename VWORKTYPE>
    int const gels_templ_driver_vec( MATA &A, VTYPE &b, VWORKTYPE &work );

    template<typename MATA, typename MATV, typename VWORKTYPE>
    int const gels_templ_driver_mat( MATA &A, MATV &b, VWORKTYPE &work );



}

int const pu::gels_work_size( double const &,
			      int const m, int const n, int const nrhs ) {
    	int info(0);
	char const transp_char('N');
	int const lwork(-1);

	double work(0);

	LAPACK_DGELS( &transp_char, &m, &n, &nrhs,
		      NULL, &m,
		      NULL, &m,
		      &work, &lwork,
		      &info );

	if ( info != 0 ) {
	    std::cerr << "gels size (double), info != 0" << std::endl;
	    std::abort();
	}

	return traits::detail::to_int(work);

}
int const pu::gels_work_size( std::complex<double> const &,
			      int const m, int const n, int const nrhs ) {
    	int info(0);
	char const transp_char('N');
	int const lwork(-1);

	traits::complex_d work(0,0);

	LAPACK_ZGELS( &transp_char, &m, &n, &nrhs,
		      NULL, &m,
		      NULL, &m,
		      traits::complex_ptr (&work), &lwork,
		      &info );

	if ( info != 0 ) {
	    std::cerr << "gels size (complex<double>), info != 0" << std::endl;
	    std::abort();
	}

	return traits::detail::to_int(work);
}

template<typename MATA, typename VTYPE, typename VWORKTYPE>
int const pu::gels_templ_driver_vec( MATA &A, VTYPE &b, VWORKTYPE &work ) {

    int const m_A( traits::matrix_size1(A) );
    int const n_A( traits::matrix_size2(A) );
    
    if ( m_A < n_A ) {
	std::cerr << "GELS: supporting only overdetermined case" << std::endl;
	std::abort();
    }

    assert( m_A == traits::vector_size( b ) );

    return
	gels_templ_diver_raw( false, m_A, n_A, 1,
			      traits::matrix_storage(A), m_A,
			      traits::vector_storage(b), m_A,
			      traits::vector_storage(work),
			      traits::vector_size(work) );
}

template<typename MATA, typename MATV, typename VWORKTYPE>
int const pu::gels_templ_driver_mat( MATA &A, MATV &b, VWORKTYPE &work ) {
    int const m_A( traits::matrix_size1(A) );
    int const n_A( traits::matrix_size2(A) );
    
    if ( m_A < n_A ) {
	std::cerr << "GELS: supporting only overdetermined case" << std::endl;
	std::abort();
    }

    assert( m_A == traits::matrix_size1( b ) );

    return
	gels_templ_diver_raw( false, m_A, n_A, traits::matrix_size2(b),
			      traits::matrix_storage(A), m_A,
			      traits::matrix_storage(b),
			      traits::matrix_size1(b),
			      traits::vector_storage(work),
			      traits::vector_size(work) );
}




int const
pu::gels( boost::numeric::ublas::matrix<double,
	  boost::numeric::ublas::column_major>
          &A,
	  boost::numeric::ublas::vector<double> &b,
	  boost::numeric::ublas::vector<double> &work ) {
    return gels_templ_driver_vec( A, b, work );
}

int const
pu::gels( boost::numeric::ublas::matrix< std::complex<double>,
	  boost::numeric::ublas::column_major >
          &A,
	  boost::numeric::ublas::vector<std::complex<double> > &b,
	  boost::numeric::ublas::vector<std::complex<double> > &work ) {
    return gels_templ_driver_vec( A, b, work );
}


int const
pu::gels( boost::numeric::ublas::matrix<double,
	  boost::numeric::ublas::column_major>
          &A,
	  boost::numeric::ublas::matrix<double,
	  boost::numeric::ublas::column_major>
	  &b,
	  boost::numeric::ublas::vector<double> &work ) {
    return gels_templ_driver_mat( A, b, work );
}

int const
pu::gels( boost::numeric::ublas::matrix< std::complex<double>,
	  boost::numeric::ublas::column_major >
          &A,
	  boost::numeric::ublas::matrix< std::complex<double>,
	  boost::numeric::ublas::column_major>
	  &b,
	  boost::numeric::ublas::vector<std::complex<double> > &work ) {
    return gels_templ_driver_mat( A, b, work );
}

